/*
 * Copyright (c) 2020 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Ecore.h>
#include <Ecore_Input.h>
#include <Ecore_Wl2.h>
#include <Evas.h>

#include <map>
#include <mutex>
#include <string>
#include <thread>

#include "media_key.h"
#include "log_private.hh"

#ifdef EXPORT_API
#undef EXPORT_API
#endif
#define EXPORT_API __attribute__((visibility("default")))

namespace {

class MediaKey {
 private:
  MediaKey() = default;
  ~MediaKey() {
    if (!disposed_)
      Dispose();
  }

 public:
  static MediaKey& GetInst() {
    static MediaKey inst;

    std::lock_guard<std::recursive_mutex> lock(inst.GetMutex());
    if (inst.disposed_)
      inst.Init();
    return inst;
  }

  void Dispose() {
    Fini();
  }

  int Reserve(media_key_event_cb cb, void* data) {
    _W("Reserve");
    std::lock_guard<std::recursive_mutex> lock(GetMutex());
    cb_ = cb;
    data_ = data;

    int ret = MEDIA_KEY_ERROR_NONE;
    if (!grabbed_)
      ret = Grab();

    return ret;
  }

  void Release() {
    _W("Release");
    std::lock_guard<std::recursive_mutex> lock(GetMutex());
    if (grabbed_)
      Ungrab();

    cb_ = nullptr;
    data_ = nullptr;
  }

 private:
  void Init() {
    key_map_ = {
      { "XF86AudioPlay", MEDIA_KEY_PLAY },
      { "XF86AudioStop", MEDIA_KEY_STOP },
      { "XF86AudioPause", MEDIA_KEY_PAUSE },
      { "XF86AudioNext", MEDIA_KEY_NEXT },
      { "XF86AudioPrev", MEDIA_KEY_PREVIOUS },
      { "XF86AudioRewind", MEDIA_KEY_REWIND },
      { "XF86AudioForward", MEDIA_KEY_FASTFORWARD },
      { "XF86AudioPlayPause", MEDIA_KEY_PLAYPAUSE },
      { "XF86AudioMedia", MEDIA_KEY_MEDIA }
    };

    ecore_wl2_init();

    if (!ecore_wl2_display_connect(NULL)) {
      _E("Failed to connect to wl2 display");
      ecore_wl2_shutdown();
      return;
    }

    disposed_ = false;
  }

  void Fini() {
    Ecore_Wl2_Display* dpy = ecore_wl2_connected_display_get(NULL);
    if (dpy)
      ecore_wl2_display_disconnect(dpy);

    ecore_wl2_shutdown();
    disposed_ = true;
  }

  int Grab() {
    for (auto i = key_map_.begin(); i != key_map_.end(); i++) {
      _D("key(%s)", i->first.c_str());
      auto ret = ecore_wl2_window_keygrab_set(NULL, i->first.c_str(),
          0, 0, 0, ECORE_WL2_WINDOW_KEYGRAB_OVERRIDE_EXCLUSIVE);
      if (ret != EINA_TRUE) {
        _E("Failed to grab key(%s)", i->first.c_str());
        Ungrab();
        return MEDIA_KEY_ERROR_OPERATION_FAILED;
      }
    }

    key_down_handler_ = ecore_event_handler_add(ECORE_EVENT_KEY_DOWN,
        MediaKeyPressCb, this);
    if (!key_down_handler_)
      _E("Failed to register key down event handler");

    key_up_handler_ = ecore_event_handler_add(ECORE_EVENT_KEY_UP,
        MediaKeyReleaseCb, this);
    if (!key_up_handler_)
      _E("Failed to register key up event handler");

    grabbed_ = true;
    return 0;
  }

  void Ungrab() {
    if (key_up_handler_) {
      ecore_event_handler_del(key_up_handler_);
      key_up_handler_ = nullptr;
    }

    if (key_down_handler_) {
      ecore_event_handler_del(key_down_handler_);
      key_down_handler_ = nullptr;
    }

    for (auto i = key_map_.rbegin(); i != key_map_.rend(); i++) {
      _D("key(%s)", i->first.c_str());
      auto ret = ecore_wl2_window_keygrab_unset(NULL, i->first.c_str(), 0, 0);
      if (ret != EINA_TRUE)
        _W("Failed to ungrab key(%s)", i->first.c_str());
    }

    grabbed_ = false;
  }

  std::recursive_mutex& GetMutex() const {
    return mutex_;
  }

  static Eina_Bool MediaKeyPressCb(void* data, int type, void* event) {
    Evas_Event_Key_Down* ev = static_cast<Evas_Event_Key_Down*>(event);
    if (ev == nullptr) {
      _E("Invalid parameter");
      return ECORE_CALLBACK_RENEW;
    }

    auto* handle = static_cast<MediaKey*>(data);
    if (handle->cb_ == nullptr)
      return ECORE_CALLBACK_RENEW;

    media_key_e media_key = handle->key_map_[ev->keyname];
    handle->cb_(media_key, MEDIA_KEY_STATUS_PRESSED, handle->data_);
    return ECORE_CALLBACK_RENEW;
  }

  static Eina_Bool MediaKeyReleaseCb(void* data, int type, void* event) {
    Evas_Event_Key_Up* ev = static_cast<Evas_Event_Key_Up*>(event);
    if (ev == nullptr) {
      _E("Invalid parameter");
      return ECORE_CALLBACK_RENEW;
    }

    auto* handle = static_cast<MediaKey*>(data);
    if (handle->cb_ == nullptr)
      return ECORE_CALLBACK_RENEW;

    media_key_e media_key = handle->key_map_[ev->keyname];
    handle->cb_(media_key, MEDIA_KEY_STATUS_RELEASED, handle->data_);
    return ECORE_CALLBACK_RENEW;
  }

 private:
  bool disposed_ = true;
  bool grabbed_ = false;
  media_key_event_cb cb_ = nullptr;
  void* data_ = nullptr;
  Ecore_Event_Handler* key_up_handler_ = nullptr;
  Ecore_Event_Handler* key_down_handler_ = nullptr;
  std::map<std::string, media_key_e> key_map_;
  mutable std::recursive_mutex mutex_;
};

}  // namespace

extern "C" EXPORT_API int media_key_reserve(media_key_event_cb callback,
    void* user_data) {
  if (callback == nullptr) {
    _E("Invalid parameter");
    return MEDIA_KEY_ERROR_INVALID_PARAMETER;
  }

  return ::MediaKey::GetInst().Reserve(callback, user_data);
}

extern "C" EXPORT_API int media_key_release(void) {
  ::MediaKey::GetInst().Release();
  return MEDIA_KEY_ERROR_NONE;
}

